#pragma once

#include "DXUT.h"

#include "reflectiveShadowMap.h"
#include "readBack.h"
#include "dxhelpers.h"
#include "sphericalharmonics.h"
#include "lpv.fx"

#include "config.h"

namespace LPV {
	struct GPULightPropagationVolume {
		static const int NUM_ITERATIONS = LPV_NUM_ITERATIONS;

		int width, height, depth;
		float cellSize; // in world coordinates
		D3DXVECTOR3 minCorner, maxCorner; // in world coordinates

		bool withOcclusion;

		struct GeometryVolume {
			ID3D10Texture3D *texture;
			ID3D10ShaderResourceView *SRV;
			ID3D10RenderTargetView *RTV;

			GeometryVolume() {
				texture = NULL;
				SRV = NULL;
				RTV = NULL;
			}

			void init( ID3D10Device *device, int width, int height, int depth ) {
				// create the 3d textures
				device->CreateTexture3D( D3D10Texture3DDesc(DXGI_FORMAT_R16G16B16A16_FLOAT, D3D10_USAGE_DEFAULT, width, height, depth ).noMipMapLevels().bindFlags( D3D10_BIND_FLAG_CPP().useAsShaderResource().useAsRenderTarget() ), NULL, &texture );
				device->CreateShaderResourceView( texture, NULL, &SRV );
				device->CreateRenderTargetView( texture, NULL, &RTV );
			}

			void clear( ID3D10Device *device ) {
				float ClearColor[4] = { 0.0f, 0.0f, 0.0f, 0.0f };

				device->ClearRenderTargetView( RTV, ClearColor );
			}

			void setRenderTarget( ID3D10Device *device ) {
				device->OMSetRenderTargets( 1, &RTV, NULL );
			}

			void setResource( ID3D10EffectShaderResourceVariable *variable ) {
				variable->SetResource( SRV );
			}

			~GeometryVolume() {
				SAFE_RELEASE( texture );
				SAFE_RELEASE( SRV );
				SAFE_RELEASE( RTV );
			}
		};

		struct LightVolume {
			ID3D10Texture3D *textures[3];
			ID3D10ShaderResourceView *SRVs[3];
			ID3D10RenderTargetView *RTVs[3];

			LightVolume() {
				for( int i = 0 ; i < 3 ; i++ ) {
					textures[i] = NULL;
					SRVs[i] = NULL;
					RTVs[i] = NULL;
				}
			}

			void init( ID3D10Device *device, int width, int height, int depth ) {
				// create the 3d textures
				for( int i = 0 ; i < 3 ; i++ ) {
					device->CreateTexture3D( D3D10Texture3DDesc(DXGI_FORMAT_R16G16B16A16_FLOAT, D3D10_USAGE_DEFAULT, width, height, depth ).noMipMapLevels().bindFlags( D3D10_BIND_FLAG_CPP().useAsShaderResource().useAsRenderTarget() ), NULL, &textures[ i ] );
					device->CreateShaderResourceView( textures[i], NULL, &SRVs[i] );
					device->CreateRenderTargetView( textures[i], NULL, &RTVs[i] );
				}
			}

			void clear( ID3D10Device *device ) {
				float ClearColor[4] = { 0.0f, 0.0f, 0.0f, 0.0f };

				for( int i = 0 ; i < 3 ; i++ ) {
					device->ClearRenderTargetView( RTVs[i], ClearColor );
				}
			}

			void setRenderTarget( ID3D10Device *device ) {
				device->OMSetRenderTargets( 3, RTVs, NULL );
			}

			void setResource( ID3D10EffectShaderResourceVariable *variable ) {
				variable->SetResourceArray( SRVs, 0, 3 );
			}

			~LightVolume() {
				for( int i = 0 ; i < 3 ; i++ ) {
					SAFE_RELEASE( textures[i] );
					SAFE_RELEASE( SRVs[i] );
					SAFE_RELEASE( RTVs[i] );
				}
			}
		};

		GeometryVolume geometryVolume;
#ifdef LPV_PING_PONG
		static const int LPV_COUNT = 2;
#else
		static const int LPV_COUNT = NUM_ITERATIONS;
#endif
		LightVolume LPVs[ LPV_COUNT ];
		LightVolume accumulatedLPV;

		EffectGPUInjection gpuInjection;
		ID3D10StateBlock *stateBlock;

		GPULightPropagationVolume() {
			stateBlock = NULL;

			withOcclusion = true;
		}

		~GPULightPropagationVolume() {
			SAFE_RELEASE( stateBlock );
		}

		void init( ID3D10Device *device, int maxResolution, const D3DXVECTOR3 &minCorner, const D3DXVECTOR3 &maxCorner ) {
			this->minCorner = minCorner;
			this->maxCorner = maxCorner;

			D3DXVECTOR3 volSize = maxCorner - minCorner;
			float maxLength = max( volSize.x, max( volSize.y, volSize.z ) );
			cellSize = maxLength / maxResolution;

			width = int( volSize.x / cellSize + 0.5f );
			height = int( volSize.y / cellSize + 0.5f );
			depth = int( volSize.z / cellSize + 0.5f );

			for( int i = 0 ; i < LPV_COUNT ; i++ ) {
				LPVs[ i ].init( device, width, height, depth );
			}
			accumulatedLPV.init( device, width, height, depth );

			geometryVolume.init( device, width, height, depth );

			gpuInjection.init( device, NULL, 0 );
			
			int size[] = { width, height, depth, 0 };
			gpuInjection.lpv_size->SetIntVector( size );
			gpuInjection.set_lpv_minCorner( minCorner );
			gpuInjection.set_lpv_cellSize( cellSize );

			D3D10_STATE_BLOCK_MASK mask;
			D3D10StateBlockMaskEnableAll( &mask );
			D3D10CreateStateBlock( device, &mask, &stateBlock );
		}

		void reloadEffect( ID3D10Device *device ) {
			gpuInjection.destroy();
			gpuInjection.init( device, NULL, 0 );

			int size[] = { width, height, depth, 0 };
			gpuInjection.lpv_size->SetIntVector( size );
			gpuInjection.set_lpv_minCorner( minCorner );
			gpuInjection.set_lpv_cellSize( cellSize );
		}

		void injectRSM( ID3D10Device *device, const ReflectiveShadowMap &rsm ) {
			stateBlock->Capture();

			LPVs[0].clear( device );
			LPVs[0].setRenderTarget( device );

			// set the viewport
			D3D10_VIEWPORT viewport;
			viewport.TopLeftX = viewport.TopLeftY = 0;
			viewport.Width = width;
			viewport.Height = height;
			viewport.MinDepth = 0.0f;
			viewport.MaxDepth = 1.0f;

			device->RSSetViewports( 1, &viewport );

			gpuInjection.set_worldPosition( rsm.renderTargetSRV[ ReflectiveShadowMap::RT_POSWORLD ] );
			gpuInjection.set_worldNormal( rsm.renderTargetSRV[ ReflectiveShadowMap::RT_NORMALWORLD ] );
			gpuInjection.set_lightFlux( rsm.renderTargetSRV[ ReflectiveShadowMap::RT_FLUX ] );

			gpuInjection.GpuInjection->GetPassByIndex(0)->Apply(0);

			device->IASetInputLayout( NULL );
			device->IASetPrimitiveTopology( D3D10_PRIMITIVE_TOPOLOGY_POINTLIST );
			device->Draw( rsm.resolution * rsm.resolution, 0 );

#if 1
			geometryVolume.clear( device );
			geometryVolume.setRenderTarget( device );

			gpuInjection.GpuOcclusionInjection->GetPassByIndex(0)->Apply(0);

			device->Draw( rsm.resolution * rsm.resolution, 0 );
#endif

			stateBlock->Apply();
		}

		void propagate( ID3D10Device *device ) {
			stateBlock->Capture();

			// set the viewport
			D3D10_VIEWPORT viewport;
			viewport.TopLeftX = viewport.TopLeftY = 0;
			viewport.Width = width;
			viewport.Height = height;
			viewport.MinDepth = 0.0f;
			viewport.MaxDepth = 1.0f;

			device->RSSetViewports( 1, &viewport );

			accumulatedLPV.clear( device );

			geometryVolume.setResource( gpuInjection.gv );

			device->IASetInputLayout( NULL );
			device->IASetPrimitiveTopology( D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST );

			for( int i = 1 ; i < NUM_ITERATIONS ; i++ ) {
#ifdef LPV_PING_PONG
				int currentLPV = i % 2;
				int previousLPV = 1 - currentLPV;
#else
				int previousLPV = i - 1;
				int currentLPV = i;
#endif

				LPVs[ previousLPV ].setResource( gpuInjection.lpv );

				// not using additive blending during propagation
				//LPVs[ i ].clear( device );

				ID3D10RenderTargetView *RTVs[] = {
					LPVs[currentLPV].RTVs[0], LPVs[currentLPV].RTVs[1], LPVs[currentLPV].RTVs[2],
					accumulatedLPV.RTVs[0], accumulatedLPV.RTVs[1], accumulatedLPV.RTVs[2],
				};
				device->OMSetRenderTargets( 6, RTVs, NULL );
				
				if( i == 1 || !withOcclusion ) {
					// no occlusion on the first iteration (to avoid self-shadowing)
					gpuInjection.GpuPropagation->GetPassByIndex(0)->Apply(0);
				}
				else {
					gpuInjection.GpuOccludedPropagation->GetPassByIndex(0)->Apply(0);
				}
				device->DrawInstanced( 3, depth, 0, 0 );
			}

			stateBlock->Apply();
		}
	};
}