//--------------------------------------------------------------------------------------
// File: MeshLoader10.h
//
// Wrapper class for ID3DX10Mesh interface. Handles loading mesh data from an .obj file
// and resource management for material textures.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//--------------------------------------------------------------------------------------
#ifndef _MESHLOADER10_H_
#define _MESHLOADER10_H_
#pragma once

#include "material.h"

#define ERROR_RESOURCE_VALUE 1

// Vertex format
struct VERTEX
{
    D3DXVECTOR3 position;
    D3DXVECTOR3 normal;
    D3DXVECTOR2 texcoord;
};


// Used for a hashtable vertex cache when creating the mesh from a .obj file
struct CacheEntry
{
    UINT index;
    CacheEntry* pNext;
};




class CMeshLoader10
{
public:
            CMeshLoader10();
            ~CMeshLoader10();

    HRESULT Create( ID3D10Device* pd3dDevice, const WCHAR* strFilename, float scale = 1.0f );

	void    Destroy();

 	HRESULT Restore( ID3D10Device* pd3dDevice, const WCHAR* strFilename );

	void Dump( const WCHAR *dstFilename );

    UINT    GetNumMaterials() const
    {
        return m_Materials.GetSize();
    }
    Material* GetMaterial( UINT iMaterial )
    {
        return m_Materials.GetAt( iMaterial );
    }

    UINT GetNumSubsets()
    {
        return m_NumAttribTableEntries;
    }
    Material* GetSubsetMaterial( UINT iSubset )
    {
        return m_Materials.GetAt( m_pAttribTable[iSubset].AttribId );
    }

    ID3DX10Mesh* GetMesh()
    {
        return m_pMesh;
    }
    WCHAR* GetMediaDirectory()
    {
        return m_strMediaDir;
    }

	const D3DXVECTOR3 & GetMinCorner() const {
		return m_vMinCorner;
	}

	const D3DXVECTOR3 & GetMaxCorner() const {
		return m_vMaxCorner;
	}

private:

    HRESULT LoadGeometryFromOBJ( const WCHAR* strFilename, float scale );
    HRESULT LoadMaterialsFromMTL( const WCHAR* strFileName );
    void    InitMaterial( Material* pMaterial );
	void	LoadMaterialTextures( ID3D10Device* pd3dDevice );
	HRESULT	CreateEncapsulatedMesh( ID3D10Device* pd3dDevice );
	HRESULT ReadDump( const WCHAR* strFilename );

    DWORD   AddVertex( UINT hash, VERTEX* pVertex );
    void    DeleteCache();

	void UpdateMinMax();

    ID3D10Device* m_pd3dDevice;    // Direct3D Device object associated with this mesh
    ID3DX10Mesh* m_pMesh;         // Encapsulated D3DX Mesh

//     CGrowableArray <CacheEntry*> m_VertexCache;   // Hashtable cache for locating duplicate vertices
    CGrowableArray <VERTEX> m_Vertices;      // Filled and copied to the vertex buffer
    CGrowableArray <DWORD> m_Indices;       // Filled and copied to the index buffer
    CGrowableArray <DWORD> m_Attributes;    // Filled and copied to the attribute buffer
    CGrowableArray <Material*> m_Materials;     // Holds material properties per subset

    UINT        m_NumAttribTableEntries;
    D3DX10_ATTRIBUTE_RANGE *m_pAttribTable;

	D3DXVECTOR3 m_vMinCorner, m_vMaxCorner;

    WCHAR   m_strMediaDir[ MAX_PATH ];               // Directory where the mesh was found
};

#endif // _MESHLOADER_H_

