//--------------------------------------------------------------------------------------
// File: reflectiveShadowMap.h
//
// (FreeBSD License)
// 
// Copyright 2010 Andreas Kirsch. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are
// permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice, this list of
// conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice, this list
// of conditions and the following disclaimer in the documentation and/or other materials
// provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY ANDREAS KIRSCH ``AS IS'' AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ANDREAS KIRSCH OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
// ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// The views and conclusions contained in the software and documentation are those of the
// authors and should not be interpreted as representing official policies, either expressed
// or implied, of Andreas Kirsch.
//--------------------------------------------------------------------------------------
#pragma once

#include "DXUT.h"
#include <vector>

#include "dxhelpers.h"

namespace LPV {
	struct ReflectiveShadowMap {
		int resolution;

		enum {
			RT_FLUX,
			RT_NORMALWORLD,
			RT_POSWORLD,
			RT_COUNT
		};

		ID3D10Texture2D *shadowMap;
		ID3D10Texture2D *renderTarget[RT_COUNT];

		// access everything as one cubemap
		ID3D10ShaderResourceView *renderTargetSRV[RT_COUNT];
		ID3D10ShaderResourceView *shadowMapSRV;

		ID3D10RenderTargetView *renderTargetRTV[RT_COUNT];
		ID3D10DepthStencilView *shadowMapDSV;

		bool initialized;

		ReflectiveShadowMap() : initialized( false ) {}

		void create( ID3D10Device *device, int resolution ) {
			HRESULT hr;

			initialized = true;
			this->resolution = resolution;

			// create the shadow map
			V( device->CreateTexture2D( 
				D3D10Texture2DDesc( DXGI_FORMAT_R32_TYPELESS, D3D10_USAGE_DEFAULT, resolution, resolution )
				.noMipMapLevels().bindFlags( D3D10_BIND_FLAG_CPP().useAsDepthStencil().useAsShaderResource() ), NULL, &shadowMap ) );

			V( device->CreateDepthStencilView( shadowMap, D3D10Texture2DDepthStencilViewDesc( DXGI_FORMAT_D32_FLOAT ), &shadowMapDSV ) );
			V( device->CreateShaderResourceView( shadowMap, D3D10Texture2DShaderResourceViewDesc( DXGI_FORMAT_R32_FLOAT ), &shadowMapSRV ) );
			
			// create the render target
			DXGI_FORMAT renderTargetFormat = DXGI_FORMAT_R16G16B16A16_FLOAT;
			for( int rtIndex = 0 ; rtIndex < RT_COUNT ; rtIndex++ ) {
				V( device->CreateTexture2D( 
					D3D10Texture2DDesc( renderTargetFormat, D3D10_USAGE_DEFAULT, resolution, resolution )
					.noMipMapLevels().bindFlags( D3D10_BIND_FLAG_CPP().useAsRenderTarget().useAsShaderResource() ), NULL, &renderTarget[ rtIndex ] ) );

				V( device->CreateRenderTargetView( renderTarget[ rtIndex ], NULL, &renderTargetRTV[ rtIndex ] ) );
				V( device->CreateShaderResourceView( renderTarget[ rtIndex ], NULL, &renderTargetSRV[ rtIndex ] ) );
			}
		}

		void destroy() {
			if( !initialized ) {
				return;
			}

			initialized = false;

			shadowMap->Release();
			shadowMapSRV->Release();
			shadowMapDSV->Release();
			for( int rtIndex = 0 ; rtIndex < RT_COUNT ; rtIndex++ ) {
				renderTarget[ rtIndex ]->Release();
				renderTargetSRV[ rtIndex ]->Release();
				renderTargetRTV[ rtIndex ]->Release();
			}
		}

		void makeTarget( ID3D10Device *device) {
			// set the render target
			device->OMSetRenderTargets( RT_COUNT, renderTargetRTV, shadowMapDSV );

			// set the viewport
			D3D10_VIEWPORT viewport;
			viewport.TopLeftX = viewport.TopLeftY = 0;
			viewport.Width = viewport.Height = resolution;
			viewport.MinDepth = 0.0f;
			viewport.MaxDepth = 1.0f;

			device->RSSetViewports( 1, &viewport );
		}

		void clearFace( ID3D10Device *device ) {
			float ClearColor[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
			
			for( int rtIndex = 0 ; rtIndex < RT_COUNT ; rtIndex++ ) {
				device->ClearRenderTargetView( renderTargetRTV[ rtIndex ], ClearColor );
			}

			// Clear the depth stencil
			device->ClearDepthStencilView( shadowMapDSV, D3D10_CLEAR_DEPTH, 1.0f, 0 );
		}

		~ReflectiveShadowMap() {
			destroy();
		}
	};
}