// Code from Darkplaces image.c! (GPL)

#include "DXUT.h"
#include "DXUTmisc.h"

#include "tgaLoader.h"

/*
=========================================================

TARGA LOADING

=========================================================
*/


typedef struct _TargaHeader
{
	unsigned char 	id_length, colormap_type, image_type;
	unsigned short	colormap_index, colormap_length;
	unsigned char	colormap_size;
	unsigned short	x_origin, y_origin, width, height;
	unsigned char	pixel_size, attributes;
}
TargaHeader;

void PrintTargaHeader(TargaHeader *t)
{
	DXUTOutputDebugString(L"TargaHeader:\nuint8 id_length = %i;\nuint8 colormap_type = %i;\nuint8 image_type = %i;\nuint16 colormap_index = %i;\nuint16 colormap_length = %i;\nuint8 colormap_size = %i;\nuint16 x_origin = %i;\nuint16 y_origin = %i;\nuint16 width = %i;\nuint16 height = %i;\nuint8 pixel_size = %i;\nuint8 attributes = %i;\n", t->id_length, t->colormap_type, t->image_type, t->colormap_index, t->colormap_length, t->colormap_size, t->x_origin, t->y_origin, t->width, t->height, t->pixel_size, t->attributes);
}

/*
=============
LoadTGA
=============
*/
HRESULT LoadTGA_BGRA (const unsigned char *f, int filesize, ImageInfo &image)
{
	int x, y, pix_inc, row_inci, runlen, alphabits;
	unsigned int *pixbufi;
	const unsigned char *fin, *enddata;
	TargaHeader targa_header;
	unsigned int palettei[256];
	union
	{
		unsigned int i;
		unsigned char b[4];
	}
	bgra;

	if (filesize < 19)
		return E_INVALIDARG;

	enddata = f + filesize;

	targa_header.id_length = f[0];
	targa_header.colormap_type = f[1];
	targa_header.image_type = f[2];

	targa_header.colormap_index = f[3] + f[4] * 256;
	targa_header.colormap_length = f[5] + f[6] * 256;
	targa_header.colormap_size = f[7];
	targa_header.x_origin = f[8] + f[9] * 256;
	targa_header.y_origin = f[10] + f[11] * 256;
	targa_header.width = image.width = f[12] + f[13] * 256;
	targa_header.height = image.height = f[14] + f[15] * 256;
	targa_header.pixel_size = f[16];
	targa_header.attributes = f[17];

	if (image.width > 4096 || image.height > 4096 || image.width <= 0 || image.height <= 0)
	{
		DXUTOutputDebugString(L"LoadTGA: invalid size\n");
		PrintTargaHeader(&targa_header);
		return E_INVALIDARG;
	}

	// advance to end of header
	fin = f + 18;

	// skip TARGA image comment (usually 0 bytes)
	fin += targa_header.id_length;

	// read/skip the colormap if present (note: according to the TARGA spec it
	// can be present even on truecolor or greyscale images, just not used by
	// the image data)
	if (targa_header.colormap_type)
	{
		if (targa_header.colormap_length > 256)
		{
			DXUTOutputDebugString(L"LoadTGA: only up to 256 colormap_length supported\n");
			PrintTargaHeader(&targa_header);
			return E_INVALIDARG;
		}
		if (targa_header.colormap_index)
		{
			DXUTOutputDebugString(L"LoadTGA: colormap_index not supported\n");
			PrintTargaHeader(&targa_header);
			return E_INVALIDARG;
		}
		if (targa_header.colormap_size == 24)
		{
			for (x = 0;x < targa_header.colormap_length;x++)
			{
				bgra.b[0] = *fin++;
				bgra.b[1] = *fin++;
				bgra.b[2] = *fin++;
				bgra.b[3] = 255;
				palettei[x] = bgra.i;
			}
		}
		else if (targa_header.colormap_size == 32)
		{
			memcpy(palettei, fin, targa_header.colormap_length*4);
			fin += targa_header.colormap_length * 4;
		}
		else
		{
			DXUTOutputDebugString(L"LoadTGA: Only 32 and 24 bit colormap_size supported\n");
			PrintTargaHeader(&targa_header);
			return E_INVALIDARG;
		}
	}

	// check our pixel_size restrictions according to image_type
	switch (targa_header.image_type & ~8)
	{
	case 2:
		if (targa_header.pixel_size != 24 && targa_header.pixel_size != 32)
		{
			DXUTOutputDebugString(L"LoadTGA: only 24bit and 32bit pixel sizes supported for type 2 and type 10 images\n");
			PrintTargaHeader(&targa_header);
			return E_INVALIDARG;
		}
		break;
	case 3:
		// set up a palette to make the loader easier
		for (x = 0;x < 256;x++)
		{
			bgra.b[0] = bgra.b[1] = bgra.b[2] = x;
			bgra.b[3] = 255;
			palettei[x] = bgra.i;
		}
		// fall through to colormap case
	case 1:
		if (targa_header.pixel_size != 8)
		{
			DXUTOutputDebugString(L"LoadTGA: only 8bit pixel size for type 1, 3, 9, and 11 images supported\n");
			PrintTargaHeader(&targa_header);
			return E_INVALIDARG;
		}
		break;
	default:
		DXUTOutputDebugString(L"LoadTGA: Only type 1, 2, 3, 9, 10, and 11 targa RGB images supported, image_type = %i\n", targa_header.image_type);
		PrintTargaHeader(&targa_header);
		return E_INVALIDARG;
	}

	if (targa_header.attributes & 0x10)
	{
		DXUTOutputDebugString(L"LoadTGA: origin must be in top left or bottom left, top right and bottom right are not supported\n");
		return E_INVALIDARG;
	}

	// number of attribute bits per pixel, we only support 0 or 8
	alphabits = targa_header.attributes & 0x0F;
	if (alphabits != 8 && alphabits != 0)
	{
		DXUTOutputDebugString(L"LoadTGA: only 0 or 8 attribute (alpha) bits supported\n");
		return E_INVALIDARG;
	}

	image.buffer = new unsigned char[image.width * image.height * 4];
	if (!image.buffer)
	{
		DXUTOutputDebugString(L"LoadTGA: not enough memory for %i by %i image\n", image.width, image.height);
		return E_INVALIDARG;
	}

	// If bit 5 of attributes isn't set, the image has been stored from bottom to top
	if ((targa_header.attributes & 0x20) == 0)
	{
		pixbufi = (unsigned int*)image.buffer + (image.height - 1)*image.width;
		row_inci = -image.width*2;
	}
	else
	{
		pixbufi = (unsigned int*)image.buffer;
		row_inci = 0;
	}

	x = 0;
	y = 0;
	pix_inc = 1;
	if ((targa_header.image_type & ~8) == 2)
		pix_inc = (targa_header.pixel_size + 7) / 8;
	switch (targa_header.image_type)
	{
	case 1: // colormapped, uncompressed
	case 3: // greyscale, uncompressed
		if (fin + image.width * image.height * pix_inc > enddata)
			break;
		for (y = 0;y < image.height;y++, pixbufi += row_inci)
			for (x = 0;x < image.width;x++)
				*pixbufi++ = palettei[*fin++];
		break;
	case 2:
		// BGR or BGRA, uncompressed
		if (fin + image.width * image.height * pix_inc > enddata)
			break;
		if (targa_header.pixel_size == 32 && alphabits)
		{
			for (y = 0;y < image.height;y++)
				memcpy(pixbufi + y * (image.width + row_inci), fin + y * image.width * pix_inc, image.width*4);
		}
		else
		{
			for (y = 0;y < image.height;y++, pixbufi += row_inci)
			{
				for (x = 0;x < image.width;x++, fin += pix_inc)
				{
					bgra.b[0] = fin[0];
					bgra.b[1] = fin[1];
					bgra.b[2] = fin[2];
					bgra.b[3] = 255;
					*pixbufi++ = bgra.i;
				}
			}
		}
		break;
	case 9: // colormapped, RLE
	case 11: // greyscale, RLE
		for (y = 0;y < image.height;y++, pixbufi += row_inci)
		{
			for (x = 0;x < image.width;)
			{
				if (fin >= enddata)
					break; // error - truncated file
				runlen = *fin++;
				if (runlen & 0x80)
				{
					// RLE - all pixels the same color
					runlen += 1 - 0x80;
					if (fin + pix_inc > enddata)
						break; // error - truncated file
					if (x + runlen > image.width)
						break; // error - line exceeds width
					bgra.i = palettei[*fin++];
					for (;runlen--;x++)
						*pixbufi++ = bgra.i;
				}
				else
				{
					// uncompressed - all pixels different color
					runlen++;
					if (fin + pix_inc * runlen > enddata)
						break; // error - truncated file
					if (x + runlen > image.width)
						break; // error - line exceeds width
					for (;runlen--;x++)
						*pixbufi++ = palettei[*fin++];
				}
			}
		}
		break;
	case 10:
		// BGR or BGRA, RLE
		if (targa_header.pixel_size == 32 && alphabits)
		{
			for (y = 0;y < image.height;y++, pixbufi += row_inci)
			{
				for (x = 0;x < image.width;)
				{
					if (fin >= enddata)
						break; // error - truncated file
					runlen = *fin++;
					if (runlen & 0x80)
					{
						// RLE - all pixels the same color
						runlen += 1 - 0x80;
						if (fin + pix_inc > enddata)
							break; // error - truncated file
						if (x + runlen > image.width)
							break; // error - line exceeds width
						bgra.b[0] = fin[0];
						bgra.b[1] = fin[1];
						bgra.b[2] = fin[2];
						bgra.b[3] = fin[3];
						fin += pix_inc;
						for (;runlen--;x++)
							*pixbufi++ = bgra.i;
					}
					else
					{
						// uncompressed - all pixels different color
						runlen++;
						if (fin + pix_inc * runlen > enddata)
							break; // error - truncated file
						if (x + runlen > image.width)
							break; // error - line exceeds width
						for (;runlen--;x++)
						{
							bgra.b[0] = fin[0];
							bgra.b[1] = fin[1];
							bgra.b[2] = fin[2];
							bgra.b[3] = fin[3];
							fin += pix_inc;
							*pixbufi++ = bgra.i;
						}
					}
				}
			}
		}
		else
		{
			for (y = 0;y < image.height;y++, pixbufi += row_inci)
			{
				for (x = 0;x < image.width;)
				{
					if (fin >= enddata)
						break; // error - truncated file
					runlen = *fin++;
					if (runlen & 0x80)
					{
						// RLE - all pixels the same color
						runlen += 1 - 0x80;
						if (fin + pix_inc > enddata)
							break; // error - truncated file
						if (x + runlen > image.width)
							break; // error - line exceeds width
						bgra.b[0] = fin[0];
						bgra.b[1] = fin[1];
						bgra.b[2] = fin[2];
						bgra.b[3] = 255;
						fin += pix_inc;
						for (;runlen--;x++)
							*pixbufi++ = bgra.i;
					}
					else
					{
						// uncompressed - all pixels different color
						runlen++;
						if (fin + pix_inc * runlen > enddata)
							break; // error - truncated file
						if (x + runlen > image.width)
							break; // error - line exceeds width
						for (;runlen--;x++)
						{
							bgra.b[0] = fin[0];
							bgra.b[1] = fin[1];
							bgra.b[2] = fin[2];
							bgra.b[3] = 255;
							fin += pix_inc;
							*pixbufi++ = bgra.i;
						}
					}
				}
			}
		}
		break;
	default:
		// unknown image_type
		break;
	}

	return S_OK;
}
